#|

 Gtk ffi

 Copyright (c) 2004 by Vasilis Margioulas <vasilism@sch.gr>

 You have the right to distribute and use this software as governed by 
 the terms of the Lisp Lesser GNU Public License (LLGPL):

    (http://opensource.franz.com/preamble.html)
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 Lisp Lesser GNU Public License for more details.
 
|#


(defpackage :gtk-ffi (:use :lisp :ffx
                       :uffi))

(in-package :gtk-ffi)

(defconstant c-null (make-null-pointer '(* :void)))
(defconstant c-null-int (make-null-pointer :int))

(defvar *gtk-debug* nil)


(defun int-slot-indexed (obj obj-type slot index)
  (declare (ignorable obj-type))
  (deref-array
   (get-slot-pointer obj obj-type slot)
   '(:array :int) index))

(defun (setf int-slot-indexed) (new-value obj obj-type slot index)
  (declare (ignorable obj-type))
  (setf (deref-array
         (get-slot-pointer obj obj-type slot)
         '(:array :int) index)
    new-value))


(eval-when (:compile-toplevel :load-toplevel :execute)
  (export '(c-null c-null-int int-slot-indexed load-gtk-libs))
  (defun gtk-function-name (lisp-name)
    (substitute #\_ #\- lisp-name))
  (defun libname (lib)
    (ecase lib
      (:gobject #+win32 "libgobject-2.0-0.dll"
		#-win32 "libgobject-2.0.so")
      (:glib #+win32 "libglib-2.0-0.dll")
      (:gthread #+win32 "libgthread-2.0-0.dll")
      (:gdk #+win32 "libgdk-win32-2.0-0.dll")
      (:gtk #+win32 "libgtk-win32-2.0-0.dll")
      #+libcellsgtk (:cgtk  "libcellsgtk.dll")))
  (defun load-gtk-libs ()
    (macrolet ((loadit (libname module) 
		       `(uffi:load-foreign-library 		 
			 (concatenate 'string cl-user::*gtk-lib-path* ,libname)
                         :force-load #+lispworks t #-lispworks nil
			 :module ,(string module))))
      #+(or win32 mswindows)
      (progn 
	(loadit "libgobject-2.0-0.dll" :gobject)
	(loadit "libglib-2.0-0.dll" :glib)
 	(loadit "libgthread-2.0-0.dll" :gthread)
	(loadit "libgdk-win32-2.0-0.dll" :gdk)
	(loadit "libgtk-win32-2.0-0.dll" :gtk)
	#+libcellsgtk(loadit "libcellsgtk.dll" :cgtk))
      #+macosx
      (progn
	(loadit "libgobject-2.0-0.dynlib" :gobject)
	(loadit "libglib-2.0-0.dynlib" :glib)
	(loadit "libgthread-2.0-0.dynlib" :gthread)
	(loadit "libgdk-win32-2.0-0.dynlib" :gdk)
	(loadit "libgtk-win32-2.0-0.dynlib" :gtk)
	#+libcellsgtk(loadit "libcellsgtk.dynlib" :cgtk))
      #-(or macosx win32 mswindows)
      (progn 
	(loadit "libgobject-2.0.so" :gobject)
	(loadit "libglib-2.0.so" :glib)
	(loadit "libgthread-2.0.so" :gthread)
	(loadit "libgdk-x11-2.0.so" :gdk)
	(loadit "libgtk-x11-2.0.so" :gtk)
	#+libcellsgtk(loadit "libcellsgtk.so" :cgtk))))
  #+cmu(load-gtk-libs)
  #+clisp(load-gtk-libs)
  (defun ffi-to-uffi-type (clisp-type)
    
    (if (consp clisp-type)
                (mapcar 'ffi-to-uffi-type clisp-type)
	      (case clisp-type
		((nil) :void)
		(uint :UNSIGNED-INT)
		(c-pointer :pointer-void)
		(c-ptr-null '*)
		(c-array-ptr '*)
		(c-ptr '*)
		(c-string :cstring)
		(sint32 :int)
		(uint32 :unsigned-int)
		(uint8 :unsigned-byte)
		(uint16 :short) ; no signed/unsigned types?
		(boolean :unsigned-int)
		(ulong :unsigned-long)
		(int :int)
		(long :long)
		(single-float :float)
		(double-float :double)
		(otherwise clisp-type))))

  (defun ffi-to-native-type (ffi-type)
    (uffi::convert-from-uffi-type
     (ffi-to-uffi-type ffi-type) :type))) ;; END eval-when

(defmacro def-gtk-function (library name &key arguments return-type 
                             (return-type-allocation :none)
                             (call-direct t))
  (declare (ignore return-type-allocation call-direct))
  
  (let* ((gtk-name$ (gtk-function-name (string-downcase (symbol-name name))))
         (gtk-name (intern (string-upcase gtk-name$))))
    (let ((arg-info
           (loop for arg in arguments
               for gsym = (gensym)
               if (eql 'c-string (cadr arg))
               collect (car arg) into arg$s
               and collect gsym into gsyms
               and collect gsym into pass-args
               else if (eql 'boolean (cadr arg))
               collect `(if ,(car arg) 1 0) into pass-args
               else if (eql 'c-pointer (cadr arg))
               collect `(or ,(car arg) c-null) into pass-args
               else collect (car arg) into pass-args
               finally (return (list (mapcar 'list gsyms arg$s)
                                 pass-args)))))
      `(progn
         (uffi:def-function (,gtk-name$ ,gtk-name)
             ,(mapcar (lambda (name-type)
                        (destructuring-bind (name type) name-type
                          (list name (ffi-to-uffi-type type))))
                arguments)
           :module ,(string library)
           :returning ,(ffi-to-uffi-type return-type))
         (defun ,name ,(mapcar 'car arguments)
           (when *gtk-debug*
             ,(unless (or (string= (symbol-name name) "GTK-EVENTS-PENDING")
                          (string= (symbol-name name) "GTK-MAIN-ITERATION-DO"))
                `(print (list ,(symbol-name name) :before ,@(mapcar 'car arguments)))))
           (prog1
               ,(let ((bodyform `(with-cstrings
                                     ,(car arg-info)
                                   (,gtk-name ,@(cadr arg-info)))))
                  (if (eql return-type 'boolean)
                      `(not (zerop ,bodyform))
		    (if (eql return-type 'c-string)
		      `(convert-from-cstring ,bodyform)
		      bodyform)
		    ))
	          
             (when *gtk-debug*
               ,(unless (or (string= (symbol-name name) "GTK-EVENTS-PENDING")
                            (string= (symbol-name name) "GTK-MAIN-ITERATION-DO"))
                  `(print (list ,(symbol-name name) :after ,@(mapcar 'car arguments)))))))
         (eval-when (compile load eval)
           (export ',name))))))

(defmacro def-gtk-lib-functions (library &rest functions)
  `(progn
     ,@(loop for function in functions collect
             (destructuring-bind (name (&rest args)
                                   &optional return-type
                                   return-type-allocation
                                   (call-direct t)) function
               `(def-gtk-function ,library ,name
                  ,@(when args `(:arguments ,args))
                  :return-type ,return-type
                  ,@(when return-type-allocation
                      `(:return-type-allocation ,return-type-allocation))
                  :call-direct ,call-direct)))))


(eval-when (:compile-toplevel :load-toplevel :execute)
  (defmacro callback-function ((&rest arguments) &optional return-type)
    (declare (ignore arguments return-type))
        `'c-pointer))



(defmacro def-c-struct (struct-name &rest fields)
  (let ((slot-defs (loop for field in fields
                         collecting (destructuring-bind (name type) field
                                      (list name
                                        (intern (string-upcase
                                                 (format nil "~a-supplied-p" name)))
                                        (ffi-to-uffi-type type))))))
    `(progn
       (uffi:def-struct ,struct-name
           ,@(loop for (name nil type) in slot-defs
                   collecting (list name type)))
       ;; --- make-<struct-name> ---
       ,(let ((obj (gensym)))
          `(defun ,(intern (string-upcase (format nil "make-~a" struct-name)))
             (&key ,@(loop for (name supplied nil) in slot-defs
                         collecting (list name nil supplied)))
             (let ((,obj (allocate-foreign-object ',struct-name)))
               ,@(loop for (name supplied nil) in slot-defs
                     collecting `(when ,supplied
                                   (setf (get-slot-value ,obj ',struct-name ',name) ,name)))
               ,obj)))

       ;; --- accessors ---
       ,@(mapcar (lambda (slot-def &aux
                           (slot-name (car slot-def))
                           (accessor (intern (format nil "~a-~a" struct-name slot-name))))
                   `(progn
                      (defun ,accessor (self)
                        (get-slot-value self ',struct-name ',slot-name))
                      (defun (setf ,accessor) (new-value self)
                        (setf (get-slot-value self ',struct-name ',slot-name)
                          new-value))))
           slot-defs))))

(def-c-struct gdk-event-button
  (type int)
  (window c-pointer)
  (send-event uint8)
  (time uint32)
  (x double-float)
  (y double-float)
  (axes (c-ptr double-float))
  (state uint)
  (button uint)
  (device c-pointer)
  (x_root double-float)
  (y_root double-float))

(def-c-struct gdk-event-key
  (type int)
  (window c-pointer)
  (send-event uint8)
  (time uint32)
  (state uint)
  (keyval uint)
  (length int)
  (string c-pointer)
  (hardware-keycode uint16)
  (group uint8))

(def-c-struct gdk-event-expose
  (type int)
  (window c-pointer)
  (send-event uint8)
  ;; This is probably wrong. alignment issues...
  (area-x int)
  (area-y int)
  (area-width int)
  (area-height int)
  (region c-pointer)
  (count int))

(def-c-struct gdk-event-motion
  (type int)
  (window c-pointer)
  (send-event uint8)
  (time int)
  (x double-float)
  (y double-float)
  (axes c-pointer)
  (state int)
  (is-hint uint16)
  (device c-pointer)
  (x-root double-float)
  (y-root double-float))

(defun event-type (event)
  (ecase event
    (-1 :nothing)
    (0 :delete)
    (1 :destroy)
    (2 :expose)
    (3 :notify) ; that is, pointer motion notify
    (4 :button_press)
    (5 :2button_press)
    (6 :3button_press)
    (7 :button_release)
    (8 :key_press)
    (9 :key_release)
    (10 :enter_notify)
    (11 :leave_notify)
    (12 :focus_change)
    (13 :configure)
    (14 :map)
    (15 :unmap)
    (16 :property_notify)
    (17 :selection_clear)
    (18 :selection_request)
    (19 :selection_notify)
    (20 :proximity_in)
    (21 :proximity_out)
    (22 :drag_enter)
    (23 :drag_leave)
    (24 :drag_motion)
    (25 :drag_status)
    (26 :drop_start)
    (27 :drop_finished)
    (28 :client_event)
    (29 :visibility_notify)
    (30 :no_expose)
    (31 :scroll)
    (32 :window_state)
    (33 :setting)))


(uffi:def-struct list-boolean
    (value :unsigned-int)
  (end :pointer-void))

(defmacro with-gtk-string ((var string) &rest body)
  `(let ((,var ,string))
     ,@body)
  #+not
  `(let ((,var (to-gtk-string ,string)))
     (unwind-protect 
           (progn ,@body)
        (g-free ,var))))

(defun value-set-function (type)
  (ecase type
    (c-string #'g-value-set-string)
    (c-pointer #'g-value-set-string)  ;; string-pointer
    (integer #'g-value-set-int)
    (single-float #'g-value-set-float)
    (double-float #'g-value-set-double)
    (boolean #'g-value-set-boolean)))

(defun value-type-as-int (type)
  (ecase type
    (c-string (* 16 4))
    (c-pointer (* 16 4)) ;; string-pointer
    (integer (* 6 4))
    (single-float (* 14 4))
    (double-float (* 15 4))
    (boolean (* 5 4))))

(def-c-struct type-val
    (type long)
  (val double-float)
  (val2 double-float))

(def-c-struct gslist
  (data c-pointer)
  (next c-pointer))

(def-c-struct gtk-tree-iter
  (stamp int)
  (user-data c-pointer)
  (user-data2 c-pointer)
  (user-data3 c-pointer))

(defmacro with-tree-iter ((iter-var) &body body)
  `(uffi:with-foreign-object (,iter-var 'gtk-tree-iter)
     (setf (get-slot-value ,iter-var 'gtk-tree-iter 'stamp) 0)
     (setf (get-slot-value ,iter-var 'gtk-tree-iter 'user-data) c-null)
     (setf (get-slot-value ,iter-var 'gtk-tree-iter 'user-data2) c-null)
     (setf (get-slot-value ,iter-var 'gtk-tree-iter 'user-data3) c-null)
     ,@body))


(eval-when (:compile-toplevel :load-toplevel :execute)
  (defun as-gtk-type-name (type)
    (ecase type
      (:string 'c-string)
      (:icon 'c-string)
      (:int 'int)
      (:long 'long)
      (:date 'float)
      (:float 'single-float)
      (:double 'double-float)
      (:boolean 'boolean)))
  (defun as-gtk-type (type)
    (ecase type
      (:string (* 16 4))
      (:icon (* 16 4))
      (:int (* 6 4))
      (:long (* 8 4))
      (:date (* 14 4))
      (:float (* 14 4))
      (:double (* 15 4))
      (:boolean (* 5 4)))))

(defun col-type-to-ffi-type (col-type)
  (cdr (assoc col-type '((:string . c-string) ;;2004:12:15-00:17 was c-pointer
                         (:icon . c-pointer)
                         (:boolean . boolean)
                         (:int . int)
                         (:long . long)
                         (:date . single-float)
                         (:float . single-float)
                         (:double . double-float)))))

(defmacro deref-pointer-runtime-typed (ptr type)
  "Returns a object pointed"
  (declare (ignorable type))
  `(deref-pointer ,ptr ,type))

(defun cast (ptr type)
  (deref-pointer-runtime-typed ptr (ffi-to-uffi-type type)))

(eval-when (compile load eval)
  (export '(uint c-pointer c-ptr-null c-array-ptr c-ptr c-string sint32 uint32 uint8 boolean
             ulong int long single-float double-float otherwise *gtk-debug*
             col-type-to-ffi-type deref-pointer-runtime-typed gtk-tree-iter)))


